#include "crypter.h"

#include <boost/test/unit_test.hpp>
#include <string>
#include <vector>
#include <iostream>

namespace
{
    const std::string PLAINTEXT("0434e3e09f49ea168c5bbf53f877ff4206923858aab7c7e1df25bc");
    const std::string SALT("jfjdslkfjslk4");
}

BOOST_AUTO_TEST_SUITE(crypter_tests)

BOOST_AUTO_TEST_CASE(crypter_GridEncryptShouldProduceCorrectOutput)
{
    const std::vector<unsigned char> plaintext(PLAINTEXT.begin(), PLAINTEXT.end());
    std::vector<unsigned char> encrypted;

    BOOST_CHECK(GridEncrypt(plaintext, encrypted));

    // Convert encrypted message to Base64 for easier verification.
    std::string encrypted_message(encrypted.begin(), encrypted.end());
    BOOST_CHECK_EQUAL(EncodeBase64(encrypted_message),
                      "sdENub7Ll+MeRKvGrzOPYuOTQ+Eqlbc1reeD+tdnQ9PcxubYwL883Anxbc7POgmm+wL8AfYEqa4MyeQbPRdL1A==");
}

BOOST_AUTO_TEST_CASE(crypter_GridDecryptShouldDecryptValidInput)
{
    const std::vector<unsigned char> encrypted =
            DecodeBase64("sdENub7Ll+MeRKvGrzOPYuOTQ+Eqlbc1reeD+tdnQ9PcxubYwL883Anxbc7POgmm+wL8AfYEqa4MyeQbPRdL1A==");

    // GridEncrypt allocates memory in destination while GridDecrypt doesn't.
    std::vector<unsigned char> plaintext(encrypted.size());
    BOOST_CHECK(GridDecrypt(encrypted, plaintext));
    const std::string decrypted_message(plaintext.begin(), plaintext.end());
    BOOST_CHECK_EQUAL(PLAINTEXT, decrypted_message);
}

BOOST_AUTO_TEST_SUITE_END()
